#ifndef ZAD2_QUAD_H
#define ZAD2_QUAD_H
#include <string>
#include <utility>

#include "../Info.h"
#include "Variable.h"
using namespace std;

static const std::string opNames[] = { "_U", "-", "!", "", "_UE",
                                       "_B", "+", "-", "*", "/", "%", "&", "|", "^", "_BE",
                                       "_C", "<", "<=", "==", "!=", ">=", ">", "_CE"};

class Quadruple : public enable_shared_from_this<Quadruple> {
public:
	class Op {
	public:
		enum OpType {
			UNARY,
			Neg, Not, Copy, UNARY_END,
			BINARY,
			Plus, Minus, Mul, Div, Mod, And, Or, Xor, BINARY_END,
			CMP,
			LT, LE, EQ, NEQ, GE, GT, CMP_END
		};
		OpType op;

		Op(OpType op = UNARY) : op(op) {};
		operator OpType() const { return op; };

		static OpType kind(OpType op) {
			if (op <= UNARY_END) return UNARY;
			if (op <= BINARY_END) return BINARY;
			return CMP;
		};
		static const std::string& name(OpType op) {
			return opNames[op];
		}
	};

	int lineno;
	int index;
	BasicBlockPtr block;
	set<VariablePtr> aliveAfter;

	virtual ~Quadruple() = default;
	virtual std::string toString() const {  return to_string(lineno) + "\t"; };
	virtual bool isFinal() const { return false; }

	virtual void useVariables() {
		auto self = shared_from_this();
		for (const auto& v : vars()) {
			if (v) v->uses.emplace_back(self);
		}
	};
	virtual vector<VariablePtr> vars() const { return {}; };
	virtual vector<VariablePtr> definitions() const { return {}; };
};

class QWriteVar : public Quadruple {
public:
	VariablePtr loc;
	explicit QWriteVar(VariablePtr loc) : Quadruple(), loc(std::move(loc)) {};

	std::string toString() const override { return Quadruple::toString() + "\t" + (loc ? (loc->name + " := ") : ""); }
	void useVariables() override {
		Quadruple::useVariables();
		if (loc) loc->writes.emplace_back(shared_from_this());
	}
	vector<VariablePtr> definitions() const override {
		auto ret = Quadruple::definitions();
		if (loc) ret.emplace_back(loc);
		return ret;
	};
};

class QLabel : public Quadruple {
public:
	string label;
	std::string toString() const override { return Quadruple::toString() + "\"" + label + "\":"; }
	explicit QLabel(std::string comment) : label(std::move(comment)) {}
};

class QAssign : public QWriteVar {
public:
	Op op;
	vector<VariablePtr> args;

	QAssign(VariablePtr loc, Op op, vector<VariablePtr> list) : QWriteVar(std::move(loc)), op(op), args(std::move(list)) {};
	QAssign(VariablePtr loc, VariablePtr l, Op op, VariablePtr r) : QWriteVar(std::move(loc)), op(op), args({std::move(l), std::move(r)}) {};
	QAssign(VariablePtr loc, Op op, VariablePtr x) : QWriteVar(std::move(loc)), op(op), args({std::move(x)}) {};

	std::string toString() const override {
		if (args.size() == 1)
			return QWriteVar::toString() + Op::name(op) + " " + args[0]->name;
		else
			return QWriteVar::toString() + args[0]->name + " " + Op::name(op) + " " + args[1]->name;
	}

	vector<VariablePtr> vars() const override {
		return args;
	}
};

class QJump : public Quadruple {
public:
	shared_ptr<QLabel> target;
	explicit QJump(shared_ptr<QLabel> target) : target(std::move(target)) {};

	std::string toString() const override {
		return Quadruple::toString() + "jump " + target->label;
	}

	virtual bool isFinal() const { return true; }
};

class QJumpCond : public QJump {
public:
	VariablePtr left;
	Op op;
	VariablePtr right;
	QJumpCond(shared_ptr<QLabel> to, VariablePtr left, Op op, VariablePtr right)
			: QJump(std::move(to)), left(std::move(left)), op(op), right(std::move(right)) {};

	QJumpCond(shared_ptr<QLabel> to, Op op, VariablePtr right)
			: QJumpCond(std::move(to), nullptr, op, std::move(right)) {};

	std::string toString() const override {
		if (!left)
			return Quadruple::toString() + "jump \"" + target->label + "\" if " + Op::name(op) + " " + right->name;
		else
			return Quadruple::toString() + "jump \"" + target->label + "\" if " + left->name + " " + Op::name(op) + " " + right->name;
	}

	bool isFinal() const override { return true; }

	vector<VariablePtr> vars() const override {
		auto ret = QJump::vars();
		if (left) ret.emplace_back(left);
		if (right) ret.emplace_back(right);
		return ret;
	}
};

class QParam : public Quadruple {
public:
	VariablePtr param;
	int num;

	QParam(VariablePtr param, int num) : param(std::move(param)), num(num) {};

	std::string toString() const override {
		if (num < 0) {
			return Quadruple::toString() + "\t" + param->name + " := param #" + to_string(-num);
		} else if (num == 0) {
			return Quadruple::toString() + "\tparam self " + param->name;
		} else {
			return Quadruple::toString() + "\tparam #" + to_string(num) + " <- " + param->name;
		}
	}

	vector<VariablePtr> vars() const override {
		auto ret = Quadruple::vars();
		if (param && num >= 0) ret.emplace_back(param);
		return ret;
	}

	void useVariables() override {
		Quadruple::useVariables();
		if (param && num < 0) param->writes.emplace_back(shared_from_this());
	}

	vector<VariablePtr> definitions() const override {
		auto ret = Quadruple::definitions();
		if (param && num < 0) ret.emplace_back(param);
		return ret;
	};
};

class QCall : public QWriteVar {
public:
	FunctionInfoPtr fn;
	int params;
	VariablePtr self;
	QCall(VariablePtr loc, FunctionInfoPtr fn, int params, VariablePtr self)
		: QWriteVar(std::move(loc)), fn(std::move(fn)), params(params), self(std::move(self)) {};

	std::string toString() const override {
		if (self) {
			return QWriteVar::toString() + "call " + fn->name + " of " + self->name + " with " + to_string(params) + " params";
		} else {
			return QWriteVar::toString() + "call " + fn->name + " with " + to_string(params) + " params";
		}
	}

	vector<VariablePtr> vars() const override {
		auto ret = QWriteVar::vars();
		if (self) ret.emplace_back(self);
		return ret;
	}
};

class QReturn : public Quadruple {
public:
	VariablePtr val;
	explicit QReturn(VariablePtr ret) : val(std::move(ret)) {};

	std::string toString() const override {
		if (val) {
			return Quadruple::toString() + "\treturn " + val->name;
		} else {
			return Quadruple::toString() + "\treturn";
		}
	}

	bool isFinal() const override { return true; }

	vector<VariablePtr> vars() const override {
		auto ret = Quadruple::vars();
		if (val) ret.emplace_back(val);
		return ret;
	}
};

class QAccess : public QWriteVar {
public:
	QAccess(VariablePtr loc, VariablePtr base, VariablePtr index, int multiplier, int offset)
		: QWriteVar(std::move(loc)), access(std::move(base), std::move(index), multiplier, offset)
	{};
	// leal    8(,%eax,4), %eax         # Arithmetic: multiply eax by 4 and add 8
	// leal    (%edx,%eax,2), %eax      # Arithmetic: multiply eax by 2 and add edx
	VariableLocation access;

	std::string toString() const override {
		auto ret = QWriteVar::toString();
		return ret + access.toString();
	}

	vector<VariablePtr> vars() const override {
		auto ret = QWriteVar::vars();
		if (access.base) ret.emplace_back(access.base);
		if (access.index) ret.emplace_back(access.index);
		return ret;
	}
};

class QWrite : public Quadruple {
public:
	QWrite(VariableLocation loc, VariablePtr val)
		: Quadruple(), loc(std::move(loc)), val(std::move(val))
	{};
	VariableLocation loc;
	VariablePtr val;

	std::string toString() const override { return Quadruple::toString() + "\t" + loc.toString() + " := " + val->name; }

	vector<VariablePtr> vars() const override {
		auto ret = Quadruple::vars();
		if (loc.base) ret.emplace_back(loc.base);
		if (loc.index) ret.emplace_back(loc.index);
		if (val) ret.emplace_back(val);
		return ret;
	}
};

class QAlloc : public QWriteVar {
public:
	size_t size;
	VariablePtr count;
	std::string virtSymbol;
	QAlloc(VariablePtr loc, size_t size, std::string virtSymbol, VariablePtr count = nullptr)
		: QWriteVar(std::move(loc)), size(size), count(std::move(count)), virtSymbol(std::move(virtSymbol)) {};

	std::string toString() const override {
		std::string ret = QWriteVar::toString() + "alloc (" + to_string(size) + ")";
		if (count) { ret += " [" + count->name + "]"; }
		if (!virtSymbol.empty()) { ret += " " + virtSymbol; }
		return ret;
	}

	vector<VariablePtr> vars() const override {
		auto ret = QWriteVar::vars();
		if (count) ret.emplace_back(count);
		return ret;
	}
};

class QPhi : public Quadruple {
public:
	vector<map<VariablePtr, VariablePtr>> phi;

	vector<VariablePtr> vars() const override {
		auto ret = Quadruple::vars();
		for (const auto& b : phi) {
			for (const auto& v : b) {
				ret.emplace_back(v.second);
			}
		}
		return ret;
	}

	void useVariables() override {
		set<VariablePtr> written;
		auto self = shared_from_this();
		for (const auto& b : phi) {
			for (const auto &v : b) {
				if (!written.contains(v.first)) {
					v.first->writes.emplace_back(self);
					written.emplace(v.first);
				}

				// QPhi does not place uses here, it is handled in RegisterAllocator
			}
		}
	}

	vector<VariablePtr> definitions() const override {
		auto ret = Quadruple::definitions();
		set<VariablePtr> written;
		for (const auto& b : phi) {
			for (const auto &v : b) {
				written.emplace(v.first);
			}
		}
		std::copy(written.begin(), written.end(), std::back_inserter(ret));
		return ret;
	};

	std::string toString() const override {
		if (phi.size() < 2) return "";
		std::string ret = Quadruple::toString() + "phi ";
		for (const auto& left : *phi.begin()) {
			auto &tg = left.first;
			ret += tg->name + " = [";
			bool f = true;
			for (auto b : phi) {
				if (!b.count(tg)) continue;
				if (f) f=false;
				else ret += ", ";
				ret += b[tg]->name;
			}
			ret += "] ";
		}
		return ret;
	}

};


#endif //ZAD2_QUAD_H
